/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Header that contains the headers for which Rust UAPI bindings
 * will be automatically generated by `bindgen`.
 *
 * Sorted alphabetically.
 */

#include <uapi/asm-generic/ioctl.h>
#include <uapi/drm/drm.h>
#include <uapi/drm/nova_drm.h>
#include <uapi/drm/panthor_drm.h>
#include <uapi/linux/android/binder.h>
#include <uapi/linux/mdio.h>
#include <uapi/linux/mii.h>
#include <uapi/linux/ethtool.h>
